package ru.dbotthepony.mc.otm.block.entity;

import net.minecraft.client.Minecraft;
import net.minecraft.core.BlockPos;
import net.minecraft.nbt.CompoundTag;
import net.minecraft.nbt.StringTag;
import net.minecraft.network.Connection;
import net.minecraft.network.protocol.PacketFlow;
import net.minecraft.network.protocol.game.ClientboundBlockEntityDataPacket;
import net.minecraft.world.entity.Entity;
import net.minecraft.world.entity.LivingEntity;
import net.minecraft.world.entity.item.ItemEntity;
import net.minecraft.world.entity.player.Player;
import net.minecraft.world.level.Level;
import net.minecraft.world.level.block.Block;
import net.minecraft.world.level.block.entity.BlockEntity;
import net.minecraft.world.level.block.state.BlockState;
import net.minecraft.world.level.levelgen.structure.BoundingBox;
import net.minecraft.world.phys.AABB;
import net.minecraft.world.phys.Vec3;
import ru.dbotthepony.mc.otm.OverdriveThatMatters;
import ru.dbotthepony.mc.otm.Registry;
import ru.dbotthepony.mc.otm.capability.MatteryCapability;
import ru.dbotthepony.mc.otm.matter.MatterRegistry;

import javax.annotation.Nullable;
import java.math.BigDecimal;

public class BlockEntityBlackHole extends BlockEntity {
	public BlockEntityBlackHole(BlockPos p_155229_, BlockState p_155230_) {
		super(Registry.BlockEntities.BLACK_HOLE, p_155229_, p_155230_);
		setMass(mass);
	}

	public static final BigDecimal NORMAL_MASS = new BigDecimal(10_000);
	private BigDecimal mass = NORMAL_MASS;
	private double gravitation_strength = 1;
	private boolean suppress_updates = true;

	public double getGravitationStrength() {
		return gravitation_strength;
	}

	public void collapse() {

	}

	public void addMass(BigDecimal mass) {
		setMass(this.mass.add(mass));
	}

	public void setMass(BigDecimal mass) {
		if (mass.compareTo(BigDecimal.ZERO) <= 0) {
			collapse();
			return;
		}

		this.mass = mass;
		setChanged();
		gravitation_strength = mass.divide(NORMAL_MASS, MatteryCapability.ROUND_RULES).doubleValue();

		if (gravitation_strength > 1) {
			gravitation_strength = Math.min(20, Math.log(gravitation_strength));
		}

		if (level != null && !level.isClientSide && !suppress_updates)
			level.sendBlockUpdated(getBlockPos(), getBlockState(), getBlockState(), Block.UPDATE_CLIENTS);

		affected_bounds = new BoundingBox(
			(int) (-30 * gravitation_strength),
			(int) (-30 * gravitation_strength),
			(int) (-30 * gravitation_strength),
			(int) (30 * gravitation_strength),
			(int) (30 * gravitation_strength),
			(int) (30 * gravitation_strength)
		).move(getBlockPos());

		affected_bounds_aabb = AABB.of(affected_bounds);
	}

	// shared functions
	public CompoundTag writeBlackHoleData(CompoundTag tag) {
		tag.putString("mass", mass.toString());
		return tag;
	}

	public void readBlackHoleData(CompoundTag tag) {
		if (tag.get("mass") instanceof StringTag str)
			setMass(new BigDecimal(str.getAsString()));
	}

	// disk io
	@Override
	public CompoundTag save(CompoundTag tag) {
		writeBlackHoleData(tag);
		return super.save(tag);
	}

	@Override
	public void load(CompoundTag tag) {
		super.load(tag);
		readBlackHoleData(tag);
	}

	// received either by game engine (from getUpdateTag on ClientLevelPacket)
	// or by onDataPacket
	@Override
	public void handleUpdateTag(CompoundTag tag) {
		super.handleUpdateTag(tag);
		readBlackHoleData(tag);
	}

	// called by game engine for ClientLevelPacket
	@Override
	public CompoundTag getUpdateTag() {
		return writeBlackHoleData(super.getUpdateTag());
	}

	// called by game engine on block updsates
	@Nullable
	@Override
	public ClientboundBlockEntityDataPacket getUpdatePacket() {
		return new ClientboundBlockEntityDataPacket(getBlockPos(), 0, writeBlackHoleData(new CompoundTag()));
	}

	// called by game engine by forge patches when ClientboundBlockEntityDataPacket is received
	@Override
	public void onDataPacket(Connection net, ClientboundBlockEntityDataPacket pkt) {
		super.onDataPacket(net, pkt);

		if (net.getReceiving() == PacketFlow.CLIENTBOUND) {
			handleUpdateTag(pkt.getTag());
		}
	}

	private BoundingBox affected_bounds;
	private AABB affected_bounds_aabb;

	public BoundingBox affectedBounds() {
		return affected_bounds;
	}

	public AABB affectedBoundsAABB() {
		return affected_bounds_aabb;
	}

	private void setDeltaMovement(Entity living, Vec3 center, double distance) {
		//final double mult = Math.min(2, (30 * this.gravitation_strength) / Math.max(1, Math.pow(distance, 2)));
		final double mult = Math.pow(1 - distance / (30 * this.gravitation_strength), 2) * this.gravitation_strength / 8;

		final var delta = living.position().vectorTo(center).normalize();
		living.setDeltaMovement(living.getDeltaMovement().add(delta.multiply(mult, mult, mult)));
	}

	public static <T extends BlockEntity> void clientTicker(Level level, BlockPos blockPos, BlockState blockState, T t) {
		if (t instanceof BlockEntityBlackHole tile) {
			var ply = Minecraft.getInstance().player;
			final var center = Vec3.atCenterOf(tile.getBlockPos());

			if (!ply.getAbilities().mayfly) {
				final double distance = ply.position().distanceTo(center);
				tile.setDeltaMovement(ply, center, distance);
			}

			for (var item : tile.level.getEntitiesOfClass(ItemEntity.class, tile.affected_bounds_aabb)) {
				final double distance = item.position().distanceTo(center);
				tile.setDeltaMovement(item, center, distance);
			}
		}
	}

	public static final BigDecimal HAWKING_MASS_LOSE_STEP = new BigDecimal("-0.1");

	public static <T extends BlockEntity> void ticker(Level level, BlockPos blockPos, BlockState blockState, T t) {
		if (t instanceof BlockEntityBlackHole tile) {
			tile.suppress_updates = false;

			final var center = Vec3.atCenterOf(tile.getBlockPos());

			for (var living : level.getEntitiesOfClass(LivingEntity.class, tile.affected_bounds_aabb)) {
				final double distance = living.position().distanceTo(center);

				if (!(living instanceof Player ply) || !ply.getAbilities().mayfly) {
					tile.setDeltaMovement(living, center, distance);
				}

				if (distance < tile.gravitation_strength + 1) {
					living.hurt(Registry.DAMAGE_EVENT_HORIZON, (float) (tile.gravitation_strength / distance));
				}
			}

			for (var item : level.getEntitiesOfClass(ItemEntity.class, tile.affected_bounds_aabb)) {
				final double distance = item.position().distanceTo(center);
				tile.setDeltaMovement(item, center, distance);

				if (distance < tile.gravitation_strength + 1) {
					if (item.hurt(Registry.DAMAGE_EVENT_HORIZON, (float) (tile.gravitation_strength / distance)) && item.isRemoved()) {
						var mass = MatterRegistry.getMatterValue(item.getItem());

						if (mass.compareTo(BigDecimal.ZERO) > 0)
							tile.addMass(mass);
					}
				}
			}

			// шанс 1% что черная дыра потеряет 0.1 MtU каждую секунду
			if (level.random.nextDouble() < 0.01 * 0.05) {
				tile.addMass(HAWKING_MASS_LOSE_STEP);
			}
		}
	}
}
